from jax import numpy as jnp
from jax import Array

from thermox.utils import (
    ProcessedDriftMatrix,
    ProcessedDiffusionMatrix,
    handle_matrix_inputs,
)
from thermox.sampler import expm_vp


def mean(
    t: float,
    x0: Array,
    A: Array | ProcessedDriftMatrix,
    b: Array,
    D: Array | ProcessedDiffusionMatrix,
) -> Array:
    """Computes the mean of p(x_t | x_0)

    For x_t evolving according to the SDE:

    dx = - A * (x - b) dt + sqrt(D) dW

    Args:
        ts: Times at which samples are collected. Includes time for x0.
        x0: Initial state of the process.
        A: Drift matrix (Array or thermox.ProcessedDriftMatrix).
            Note: If a thermox.ProcessedDriftMatrix instance is used as input,
            must be the transformed drift matrix, A_y, given by thermox.preprocess,
            not thermox.utils.preprocess_drift_matrix.
        b: Drift displacement vector.
        D: Diffusion matrix (Array or thermox.ProcessedDiffusionMatrix).

    """
    A_y, D = handle_matrix_inputs(A, D)

    y0 = D.sqrt_inv @ (x0 - b)
    return b + D.sqrt @ expm_vp(A_y, y0, t)


def covariance(
    t: float,
    A: Array | ProcessedDriftMatrix,
    D: Array | ProcessedDiffusionMatrix,
) -> Array:
    """Computes the covariance of p(x_t | x_0)

    For x evolving according to the SDE:

    dx = - A * (x - b) dt + sqrt(D) dW

    Args:
        ts: Times at which samples are collected. Includes time for x0.
        A: Drift matrix (Array or thermox.ProcessedDriftMatrix).
            Note: If a thermox.ProcessedDriftMatrix instance is used as input,
            must be the transformed drift matrix, A_y, given by thermox.preprocess,
            not thermox.utils.preprocess_drift_matrix.
        D: Diffusion matrix (Array or thermox.ProcessedDiffusionMatrix).
    """
    A_y, D = handle_matrix_inputs(A, D)

    identity_diffusion_cov = (
        A_y.sym_eigvecs
        @ jnp.diag((1 - jnp.exp(-2 * A_y.sym_eigvals * t)) / (2 * A_y.sym_eigvals))
        @ A_y.sym_eigvecs.T
    )
    return D.sqrt @ identity_diffusion_cov @ D.sqrt.T


def mean_and_covariance(
    t: float,
    x0: Array,
    A: Array | ProcessedDriftMatrix,
    b: Array,
    D: Array | ProcessedDiffusionMatrix,
) -> tuple[Array, Array]:
    """Computes the mean and covariance of p(x_t | x_0)

    For x evolving according to the SDE:

    dx = - A * (x - b) dt + sqrt(D) dW

    Args:
        ts: Times at which samples are collected. Includes time for x0.
        x0: Initial state of the process.
        A: Drift matrix (Array or thermox.ProcessedDriftMatrix).
            Note: If a thermox.ProcessedDriftMatrix instance is used as input,
            must be the transformed drift matrix, A_y, given by thermox.preprocess,
            not thermox.utils.preprocess_drift_matrix.
        b: Drift displacement vector.
        D: Diffusion matrix (Array or thermox.ProcessedDiffusionMatrix).

    """
    A, D = handle_matrix_inputs(A, D)
    mean_val = mean(t, x0, A, b, D)
    covariance_val = covariance(t, A, D)
    return mean_val, covariance_val



import jax
from jax import numpy as jnp

import thermox


def test_sample_array_input():
    jax.config.update("jax_enable_x64", True)
    key = jax.random.PRNGKey(0)
    dim = 2
    dt = 0.1
    ts = jnp.arange(0, 10_000, dt)

    # Add some noise to the time points to make the timesteps different
    ts += jax.random.uniform(key, (ts.shape[0],)) * dt
    ts = ts.sort()

    A = jnp.array([[3, 2.5], [2, 4.0]])
    b = jax.random.normal(jax.random.PRNGKey(1), (dim,))
    x0 = jax.random.normal(jax.random.PRNGKey(2), (dim,))
    D = 2 * jnp.eye(dim)

    samples = thermox.sample(key, ts, x0, A, b, D, associative_scan=False)

    samp_cov = jnp.cov(samples.T)
    samp_mean = jnp.mean(samples.T, axis=1)
    assert jnp.allclose(A @ samp_cov, jnp.eye(2), atol=1e-1)
    assert jnp.allclose(samp_mean, b, atol=1e-1)

    samples_as = thermox.sample(key, ts, x0, A, b, D, associative_scan=True)
    assert jnp.allclose(samples, samples_as, atol=1e-6)



import jax
from jax import numpy as jnp

import thermox


def test_mean_and_cov():
    jax.config.update("jax_enable_x64", True)
    dim = 2
    t = 1.0

    A = jnp.array([[3, 2.5], [2, 4.0]])
    b = jax.random.normal(jax.random.PRNGKey(1), (dim,))
    x0 = jax.random.normal(jax.random.PRNGKey(2), (dim,))
    D = 2 * jnp.eye(dim)

    mean = thermox.conditional.mean(t, x0, A, b, D)
    samples = jax.vmap(
        lambda k: thermox.sample(k, jnp.array([0.0, t]), x0, A, b, D)[-1]
    )(jax.random.split(jax.random.PRNGKey(0), 1000000))
    assert mean.shape == (dim,)
    assert jnp.allclose(mean, jnp.mean(samples, axis=0), atol=1e-2)

    cov = thermox.conditional.covariance(t, A, D)
    assert cov.shape == (dim, dim)
    assert jnp.allclose(cov, jnp.cov(samples.T), atol=1e-3)

    mean_and_cov = thermox.conditional.mean_and_covariance(t, x0, A, b, D)
    assert mean_and_cov[0].shape == (dim,)
    assert mean_and_cov[1].shape == (dim, dim)
    assert jnp.allclose(mean_and_cov[0], mean, atol=1e-5)
    assert jnp.allclose(mean_and_cov[1], cov, atol=1e-5)



from functools import partial
import jax
import jax.numpy as jnp
from jax import Array

from thermox.utils import (
    handle_matrix_inputs,
    preprocess_drift_matrix,
    ProcessedDriftMatrix,
    ProcessedDiffusionMatrix,
)


def sample(
    key: Array,
    ts: Array,
    x0: Array,
    A: Array | ProcessedDriftMatrix,
    b: Array,
    D: Array | ProcessedDiffusionMatrix,
    associative_scan: bool = True,
) -> Array:
    """Collects samples from the Ornstein-Uhlenbeck process, defined as:

    dx = - A * (x - b) dt + sqrt(D) dW

    by using exact diagonalization.

    Preprocessing (diagonalization) costs O(d^3) and sampling costs O(T * d^2),
    where T=len(ts).

    If associative_scan=True then jax.lax.associative_scan is used which will run in
    time O((T/p + log(T)) * d^2) on a GPU/TPU with p cores, still with
    O(d^3) preprocessing.

    By default, this function does the preprocessing on A and D before the evaluation.
    However, the preprocessing can be done externally using thermox.preprocess
    the output of which can be used as A and D here, this will skip the preprocessing.

    Args:
        key: Jax PRNGKey.
        ts: Times at which samples are collected. Includes time for x0.
        x0: Initial state of the process.
        A: Drift matrix (Array or thermox.ProcessedDriftMatrix).
            Note: If a thermox.ProcessedDriftMatrix instance is used as input,
            must be the transformed drift matrix, A_y, given by thermox.preprocess,
            not thermox.utils.preprocess_drift_matrix.
        b: Drift displacement vector.
        D: Diffusion matrix (Array or thermox.ProcessedDiffusionMatrix).
        associative_scan: If True, uses jax.lax.associative_scan.

    Returns:
        Array-like, desired samples.
            shape: (len(ts), ) + x0.shape
    """
    A_y, D = handle_matrix_inputs(A, D)

    y0 = D.sqrt_inv @ x0
    b_y = D.sqrt_inv @ b
    ys = sample_identity_diffusion(key, ts, y0, A_y, b_y, associative_scan)
    return jax.vmap(jnp.matmul, in_axes=(None, 0))(D.sqrt, ys)


def sample_identity_diffusion(
    key: Array,
    ts: Array,
    x0: Array,
    A: Array | ProcessedDriftMatrix,
    b: Array,
    associative_scan: bool = True,
) -> Array:
    if associative_scan:
        return _sample_identity_diffusion_associative_scan(key, ts, x0, A, b)
    else:
        return _sample_identity_diffusion_scan(key, ts, x0, A, b)


def expm_vp(A, v, dt):
    out = A.eigvecs_inv @ v
    out = jnp.exp(-A.eigvals * dt) * out
    out = A.eigvecs @ out
    return out.real


def transition_cov_sqrt_vp(A, v, dt):
    diag = ((1 - jnp.exp(-2 * A.sym_eigvals * dt)) / (2 * A.sym_eigvals)) ** 0.5
    out = diag * v
    out = A.sym_eigvecs @ out
    return out.real


def _sample_identity_diffusion_scan(
    key: Array,
    ts: Array,
    x0: Array,
    A: Array | ProcessedDriftMatrix,
    b: Array,
) -> Array:
    if isinstance(A, Array):
        A = preprocess_drift_matrix(A)

    def transition_mean(x, dt):
        return b + expm_vp(A, x - b, dt)

    def next_x(x, dt, rv):
        return transition_mean(x, dt) + transition_cov_sqrt_vp(A, rv, dt)

    def scan_body(carry, dt_and_rv):
        x = carry
        dt, rv = dt_and_rv
        new_x = next_x(x, dt, rv)
        return new_x, new_x

    dts = jnp.diff(ts)
    gauss_samps = jax.random.normal(key, (len(dts),) + x0.shape)

    # Stack dts and gauss_samps along a new axis
    dt_and_rv = (dts, gauss_samps)

    _, xs = jax.lax.scan(scan_body, x0, dt_and_rv)
    xs = jnp.concatenate([jnp.expand_dims(x0, axis=0), xs], axis=0)
    return xs


def _sample_identity_diffusion_associative_scan(
    key: Array,
    ts: Array,
    x0: Array,
    A: Array | ProcessedDriftMatrix,
    b: Array,
) -> Array:
    if isinstance(A, Array):
        A = preprocess_drift_matrix(A)

    dts = jnp.diff(ts)

    # transition_mean(x, dt) = b + expm_vp(A, x - b, dt)

    gauss_samps = jax.random.normal(key, (len(dts),) + x0.shape)
    noise_terms = jax.vmap(lambda v, dt: transition_cov_sqrt_vp(A, v, dt))(
        gauss_samps, dts
    )

    @partial(jax.vmap, in_axes=(0, 0))
    def binary_associative_operator(elem_a, elem_b):
        t_a, x_a = elem_a
        t_b, x_b = elem_b
        return t_a + t_b, expm_vp(A, x_a, t_b) + x_b

    scan_times = jnp.concatenate([ts[:1], dts], dtype=float)  # [t0, dt1, dt2, ...]
    scan_input_values = jnp.concatenate(
        [x0[None] - b, noise_terms], axis=0
    )  # Shift input by b
    scan_elems = (scan_times, scan_input_values)

    scan_output = jax.lax.associative_scan(binary_associative_operator, scan_elems)
    return scan_output[1] + b  # Shift back by b
